from typing import Union

import numpy as np


def numpy_videoframe_has_alpha_pixels(
    frame: 'np.ndarray'
) -> bool:
    """
    Check if the provided numpy array has alpha
    pixels or not. If the provided numpy array
    doesn't include an alpha layer the result
    will be False.
    """
    return (
        frame.ndim == 3 and
        frame.shape[2] > 3 and
        np.any(frame[..., 3] < 255)
    )

def videoframe_has_alpha_layer(
    frame: 'VideoFrame'
) -> bool:
    """
    Check if the provided 'frame' pyav VideoFrame
    has alpha layer or not, that is defined by an
    'a' in the frame format name. This doesn't 
    mean that some pixel is transparent, it only
    means that there is an alpha layer, but the
    frame could be completely opaque.

    The code:
    - `'a' in frame.format.name`
    """
    return 'a' in frame.format.name

def update_transparency(
    transparency: Union[float, None],
    format: Union[str, None] = None
) -> Union[float, None]:
    """
    Update the provided 'transparency' value,
    that must be a value in the [0.0, 1.0] range,
    if necessary.

    If the 'format' is provided:
    - Alpha format will force an opaque alpha
    transparency if it is None
    - Non-alpha format will force a None value
    if transparency is provided
    """
    return (
        # Alpha format, no transparency => opaque
        0.0
        if (
            transparency is None and
            format is not None and
            'a' in format 
        ) else
        # Non-alpha format but transparency => None
        None
        if (
            transparency is not None and
            format is not None and
            'a' not in format
        ) else
        # Lower than 0.0 limit
        0.0
        if (
            transparency is not None and
            transparency < 0.0
        ) else
        # Greater than 1.0 limit
        1.0
        if (
            transparency is not None and
            transparency > 1.0
        ) else
        transparency
    )


"""
The video frames must be built using the
(height, width) size when giving the numpy
array that will be used for it. We will
receive the values as (width, height) but
we will invert them when needed.

The frames that come from an empty part
are flagged with the .metadata attribute
'is_from_empty_part' so we can recognize
them and ignore when combining on the
timeline. We have that metadata in the
wrapper class we created.

TODO: Check because we have a similar
module in other project or projects.
"""
from yta_editor.utils.alpha import update_transparency
from av.video.frame import VideoFrame
from av.audio.frame import AudioFrame
from av.audio.layout import AudioLayout
from typing import Union

import numpy as np


class _FrameGenerator:
    """
    Class to generate frames as numpy arrays.
    """

    # TODO: I have some library doing this with
    # colors and numpy frames, so please refactor
    def full_black(
        self,
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        transparency: Union[float, None] = None
    ) -> np.ndarray:
        """
        Get a numpy array that represents a full
        black frame of the given 'size' and with
        the given 'dtype'.

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return self._get_numpy_array(
            color = (0, 0, 0),
            size = size,
            dtype = dtype,
            transparency = transparency
        )
    
    def full_white(
        self,
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        transparency: Union[float, None] = None
    ) -> np.ndarray:
        """
        Get a numpy array that represents a full
        black frame of the given 'size' and with
        the given 'dtype'.

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return self._get_numpy_array(
            color = (255, 255, 255),
            size = size,
            dtype = dtype,
            transparency = transparency
        )

    def full_red(
        self,
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        transparency: Union[float, None] = None
    ) -> np.ndarray:
        """
        Get a numpy array that represents a full
        red frame of the given 'size' and with
        the given 'dtype'.
        
        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return self._get_numpy_array(
            color = (255, 0, 0),
            size = size,
            dtype = dtype,
            transparency = transparency
        )
    
    def full_green(
        self,
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        transparency: Union[float, None] = None
    ) -> np.ndarray:
        """
        Get a numpy array that represents a full
        green frame of the given 'size' and with
        the given 'dtype'.
        
        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return self._get_numpy_array(
            color = (0, 255, 0),
            size = size,
            dtype = dtype,
            transparency = transparency
        )
    
    def full_blue(
        self,
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        transparency: Union[float, None] = None
    ) -> np.ndarray:
        """
        Get a numpy array that represents a full
        blue frame of the given 'size' and with
        the given 'dtype'.
        
        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return self._get_numpy_array(
            color = (0, 0, 255),
            size = size,
            dtype = dtype,
            transparency = transparency
        )
    
    def _get_numpy_array(
        self,
        color: tuple[int, int, int],
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        transparency: Union[float, None] = None
    ) -> np.ndarray:
        """
        *For internal use only*

        Get the numpy array with the provided
        color and all the attributes set.

        The size must be (w, h).

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        # Transparency in [0.0, 1.0] range
        transparency = update_transparency(transparency = transparency)

        """
        We handle transparency as a float value
        in the [0.0, 1.0] range, which is easy
        conceptually, but the RGBA frames are
        made different so we need to invert the
        values:
        - Full RGBA opaque value: 255
        - Full RGBA transparent value: 0
        """

        # Transparency to [0, 255] range where
        # 0 is transparent, 255 is opaque
        transparency = (
            int(round((1.0 - transparency) * 255))
            if transparency is not None else
            transparency
        )

        dimensions = (
            4
            if transparency is not None else
            3
        )

        fill_value = (
            color + (transparency,)
            if transparency is not None else
            color
        )

        return np.full(
            shape = (size[1], size[0], dimensions),
            fill_value = fill_value,
            dtype = dtype
        )

class _BackgroundFrameGenerator:
    """
    Internal class to simplify the way we
    access to the generation of background
    frames form the general generator class.
    """

    def __init__(
        self
    ):
        self._frame_generator: _FrameGenerator = _FrameGenerator()
        """
        Shortcut to the FrameGenerator.
        """

    def full_black(
        self,
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        # TODO: I disable format by now because I
        # make it dynamic according to the 
        # 'transparency' value provided
        #format: str = 'rgb24',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        transparency: Union[float, None] = None
    ) -> VideoFrame:
        """
        Get a video frame that is completely black
        and of the given 'size'.

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return self._full_color(
            color = 'black',
            size = size,
            dtype = dtype,
            pts = pts,
            time_base = time_base,
            transparency = transparency
        )
    
    def full_white(
        self,
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        # TODO: I disable format by now because I
        # make it dynamic according to the 
        # 'transparency' value provided
        #format: str = 'rgb24',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        transparency: Union[float, None] = None
    ) -> VideoFrame:
        """
        Get a video frame that is completely white
        and of the given 'size'.

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return self._full_color(
            color = 'white',
            size = size,
            dtype = dtype,
            pts = pts,
            time_base = time_base,
            transparency = transparency
        )

    def full_red(
        self,
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        # TODO: I disable format by now because I
        # make it dynamic according to the 
        # 'transparency' value provided
        #format: str = 'rgb24',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        transparency: Union[float, None] = None
    ) -> VideoFrame:
        """
        Get a video frame that is completely red
        and of the given 'size'.

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return self._full_color(
            color = 'red',
            size = size,
            dtype = dtype,
            pts = pts,
            time_base = time_base,
            transparency = transparency
        )
    
    # TODO: Refactor all this please
    def full_green(
        self,
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        # TODO: I disable format by now because I
        # make it dynamic according to the 
        # 'transparency' value provided
        #format: str = 'rgb24',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        transparency: Union[float, None] = None
    ) -> VideoFrame:
        """
        Get a video frame that is completely green
        and of the given 'size'.

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return self._full_color(
            color = 'green',
            size = size,
            dtype = dtype,
            pts = pts,
            time_base = time_base,
            transparency = transparency
        )

    def full_blue(
        self,
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        # TODO: I disable format by now because I
        # make it dynamic according to the 
        # 'transparency' value provided
        #format: str = 'rgb24',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        transparency: Union[float, None] = None
    ) -> VideoFrame:
        """
        Get a video frame that is completely blue
        and of the given 'size'.

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return self._full_color(
            color = 'blue',
            size = size,
            dtype = dtype,
            pts = pts,
            time_base = time_base,
            transparency = transparency
        )
    
    def _full_color(
        self,
        color: str,
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        # TODO: I disable format by now because I
        # make it dynamic according to the 
        # 'transparency' value provided
        #format: str = 'rgb24',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        transparency: Union[float, None] = None
    ):
        """
        *For internal use only*

        Method that dynamically calls the internal
        method to generate the numpy according to
        the 'color' provided, that must fit the
        name of an existing method of the 
        '_FrameGenerator' instance.
        """
        format = (
            'rgba'
            if transparency is not None else
            'rgb24'
        )

        transparency = update_transparency(
            transparency = transparency,
            format = format
        )

        generator_method = getattr(self._frame_generator, f'full_{color}')

        frame = generator_method(
            size = size,
            dtype = dtype,
            transparency = transparency
        )

        return numpy_to_video_frame(
            frame = frame,
            format = format,
            pts = pts,
            time_base = time_base
        )

class VideoFrameGenerator:
    """
    Class to wrap the functionality related to
    generating a pyav video frame.

    This class is useful when we need to 
    generate the black background for empty
    parts within the tracks and in other 
    situations.
    """

    def __init__(
        self
    ):
        self.background = _BackgroundFrameGenerator()
        """
        Shortcut to the background creation.
        """

def numpy_to_video_frame(
    frame: np.ndarray,
    format: str = 'rgb24',
    pts: Union[int, None] = None,
    time_base: Union['Fraction', None] = None
) -> VideoFrame:
    """
    Transform the given numpy 'frame' into a
    pyav video frame with the given 'format'
    and also the 'pts' and/or 'time_base' if
    provided.
    """
    frame = VideoFrame.from_ndarray(
        # TODO: What if we want alpha (?)
        array = frame,
        format = format
    )

    if pts is not None:
        frame.pts = pts

    if time_base is not None:
        frame.time_base = time_base

    return frame

class AudioFrameGenerator:
    """
    Class to wrap the functionality related to
    generating a pyav audio frame.

    This class is useful when we need to 
    generate the silent audio for empty parts
    within the tracks and in other situations.
    """

    def silent(
        self,
        sample_rate: int,
        layout = 'stereo',
        number_of_samples: int = 1024,
        format = 's16',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None
    ) -> AudioFrame:
        """
        Get an audio frame that is completely silent.
        This is useful when we want to fill the empty
        parts of our tracks.
        """
        dtype = audio_format_to_dtype(format)

        if dtype is None:
            raise Exception(f'The format "{format}" is not accepted.')

        # TODO: Is this raising exception if the
        # 'layout' is not valid? I think yes (?)
        number_of_channels = len(AudioLayout(layout).channels)

        # TODO: I leave these comments below because
        # I'm not sure what is true and what is not
        # so, until it is more clear... here it is:
        # For packed (or planar) formats we apply:
        # (1, samples * channels). This is the same
        # amount of data but planar, in 1D only
        # TODO: This wasn't in the previous version
        # and it was working, we were sending the
        # same 'number_of_samples' even when 'fltp'
        # that includes the 'p'
        # TODO: This is making the audio last 2x
        # if 'p' in format:
        #     number_of_samples *= number_of_channels

        silent_numpy_array = np.zeros(
            shape = (number_of_channels, number_of_samples),
            dtype = dtype
        )

        return numpy_to_audio_frame(
            frame = silent_numpy_array,
            sample_rate = sample_rate,
            layout = layout,
            format = format,
            pts = pts,
            time_base = time_base
        )
    
def numpy_to_audio_frame(
    frame: np.ndarray,
    sample_rate: int,
    layout: str = 'stereo',
    format: str = ' s16',
    pts: Union[int, None] = None,
    time_base: Union['Fraction', None] = None
) -> AudioFrame:
    """
    Transform the given numpy 'frame' into a
    pyav audio frame with the given 'sample_rate',
    'layout' and 'format, and also the 'pts
    and/or 'time_base' if provided.
    """
    frame = AudioFrame.from_ndarray(
        array = frame,
        format = format,
        layout = layout
    )

    frame.sample_rate = sample_rate

    if pts is not None:
        frame.pts = pts

    if time_base is not None:
        frame.time_base = time_base

    return frame

# TODO: Maybe transform into a Enum (?)
def audio_format_to_dtype(
    audio_format: str
) -> Union[np.dtype, None]:
    """
    Transform the given 'audio_format' into
    the corresponding numpy dtype value. If
    the 'audio_format' is not accepted this
    method will return None.

    This method must be used when we are
    building the numpy array that will be 
    used to build a pyav audio frame because
    the pyav 'audio_format' need a specific
    np.dtype to be built.

    For example, 's16' will return 'np.int16'
    and 'fltp' will return 'np.float32'.
    """
    return {
        's16': np.int16,
        'flt': np.float32,
        'fltp': np.float32
    }.get(audio_format, None)
    


from yta_editor.utils.frame_generator import AudioFrameGenerator
from yta_editor.utils.frame_wrapper import AudioFrameWrapped
from yta_video_frame_time.t_fraction import fps_to_time_base
from yta_video_opengl.effects import EffectsStack
from yta_validation import PythonValidator
from yta_validation.parameter import ParameterValidator
from av.video.frame import VideoFrame
from av.audio.frame import AudioFrame
from quicktions import Fraction
from typing import Union


def audio_frames_and_remainder_per_video_frame(
    # TODO: Maybe force 'fps' as int (?)
    video_fps: Union[float, Fraction],
    sample_rate: int, # audio_fps
    number_of_samples_per_audio_frame: int
) -> tuple[int, int]:
    """
    Get how many full silent audio frames we
    need and the remainder for the last one
    (that could be not complete), according
    to the parameters provided.

    This method returns a tuple containing
    the number of full silent audio frames
    we need and the number of samples we need
    in the last non-full audio frame.
    """
    # Video frame duration (in seconds)
    time_base = fps_to_time_base(video_fps)
    sample_rate = Fraction(int(sample_rate), 1)

    # Example:
    # 44_100 / 60 = 735  ->  This means that we
    # will have 735 samples of sound per each
    # video frame
    # The amount of samples per frame is actually
    # the amount of samples we need, because we
    # are generating it...
    samples_per_frame = sample_rate * time_base
    # The 'nb_samples' is the amount of samples
    # we are including on each audio frame
    full_audio_frames_needed = samples_per_frame // number_of_samples_per_audio_frame
    remainder = samples_per_frame % number_of_samples_per_audio_frame
    
    return int(full_audio_frames_needed), int(remainder)

"""
These methods below are shared by the
Audio and Video class that handle and
wrap an audio or video.
"""
def apply_video_effects_to_frame_at_t(
    effects_stack: EffectsStack,
    frame: VideoFrame,
    t: Union[int, float, 'Fraction']
) -> Union[VideoFrame, 'ndarray']:
    """
    Apply the video effects to the given
    'frame' on the 't' time moment provided.

    This method should be called before 
    yielding any frame.
    """
    ParameterValidator.validate_mandatory_instance_of('frame', frame, VideoFrame)
    ParameterValidator.validate_mandatory_instance_of('effects_stack', effects_stack, EffectsStack)

    if len(effects_stack.video_effects) == 0:
        return frame

    # TODO: I think this has to preserve the
    # transparency if any of the video frames
    # that are passed to this method (from 
    # the different layers) have an alpha
    # channel or we will lose it in the
    # process... By now I'm just forcing the
    # format that comes. Or maybe not,
    # because when a frame comes and has no
    # transparency, the next frames will be
    # not visible (if coming ordered by
    # priority)... Well... maybe I'm confused
    # right now and this just modifies a 
    # single frame or a frame that has been
    # combined previously, so no worries...

    # We handle frames as 'rgb24' or 'rgba' and
    # then reformat to the expected format (if
    # needed)
    temp_format = (
        'rgba'
        if (
            frame.format.name != 'rgba' and
            'a' in frame.format.name
        ) else
        'rgb24'
        if (
            frame.format.name != 'rgb24' and
            'a' not in frame.format.name
        ) else
        frame.format.name
    )

    # Need to send the frame as a numpy for
    # the effects
    new_frame = effects_stack.apply_video_effects_at_t(
        frame = frame.to_ndarray(
            format = temp_format
        ),
        # The 't' here is the internal valid one
        t = t
    )

    """
    When applying the video effects we use
    opengl textures that return, always, an
    alpha channel, so we need to remove it
    if we don't actually need it
    """
    new_frame = (
        remove_alpha_channel(new_frame)
        if temp_format == 'rgb24' else
        new_frame
    )

    # Rebuild the VideoFrame
    new_frame = VideoFrame.from_ndarray(
        array = new_frame,
        format = temp_format
    )

    new_frame = (
        new_frame.reformat(format = frame.format.name)
        if frame.format.name != temp_format else
        new_frame
    )

    """
    We need 'time_base' and 'pts' values to
    be identified by pyav as valid frames 
    but we don't actually care about the
    values because the Timeline that renders
    will overwrite them
    """
    new_frame.time_base = (
        frame.time_base
        if frame.time_base is not None else
        Fraction(1, 60)
    )

    new_frame.pts = (
        frame.pts
        if frame.pts is not None else
        0
    )

    return new_frame

def apply_audio_effects_to_frame_at_t(
    effects_stack: EffectsStack,
    frame: Union['AudioFrame', 'ndarray'],
    t: Union[int, float, 'Fraction']
) -> Union['AudioFrame', 'ndarray']:
    """
    Apply the audio effects to the given
    'frame' on the 't' time moment provided.

    This method should be called before 
    yielding any frame.
    """
    # TODO: I think we shouldn't receive a
    # 'ndarray' here, it must be AudioFrame
    ParameterValidator.validate_mandatory_instance_of('frame', frame, [AudioFrame, 'ndarray'])

    # Need the frame as a numpy
    new_frame = (
        frame.to_ndarray()
        if PythonValidator.is_instance_of(frame, AudioFrame) else
        frame
    )
    
    new_frame = effects_stack.apply_audio_effects_at_t(
        frame = new_frame,
        # The 't' here is the internal valid one
        t = t
    )

    # Rebuild the AudioFrame
    new_frame = AudioFrame.from_ndarray(
        array = new_frame,
        format = frame.format,
        layout = frame.layout
    )

    new_frame.sample_rate = frame.sample_rate

    """
    When applying the video effects we use
    opengl textures that return, always, an
    alpha channel, so we need to remove it
    if we don't actually need it
    """
    new_frame.time_base = (
        frame.time_base
        if frame.time_base is not None else
        Fraction(1, 60)
    )

    new_frame.pts = (
        frame.pts
        if frame.pts is not None else
        0
    )

    return new_frame


# TODO: Is this method here ok (?)
def generate_silent_frames(
    fps: int,
    audio_fps: int,
    audio_samples_per_frame: int,
    layout: str = 'stereo',
    format: str = 'fltp'
) -> list[AudioFrameWrapped]:
    """
    Get the audio silent frames we need for
    a video with the given 'fps', 'audio_fps'
    and 'audio_samples_per_frame', using the
    also provided 'layout' and 'format' for
    the audio frames.

    This method is used when we have empty
    parts on our tracks and we need to 
    provide the frames, that are passed as
    AudioFrameWrapped instances and tagged as
    coming from empty parts.
    """
    audio_frame_generator: AudioFrameGenerator = AudioFrameGenerator()

    # Check how many full and partial silent
    # audio frames we need
    number_of_frames, number_of_remaining_samples = audio_frames_and_remainder_per_video_frame(
        video_fps = fps,
        sample_rate = audio_fps,
        number_of_samples_per_audio_frame = audio_samples_per_frame
    )

    # The complete silent frames we need
    silent_frame = audio_frame_generator.silent(
        sample_rate = audio_fps,
        layout = layout,
        number_of_samples = audio_samples_per_frame,
        format = format,
        pts = None,
        time_base = None
    )
    
    frames = (
        [
            AudioFrameWrapped(
                frame = silent_frame,
                is_from_empty_part = True
            )
        ] * number_of_frames
        if number_of_frames > 0 else
        []
    )

    # The remaining partial silent frames we need
    if number_of_remaining_samples > 0:
        silent_frame = audio_frame_generator.silent(
            sample_rate = audio_fps,
            # TODO: Check where do we get this value from
            layout = layout,
            number_of_samples = number_of_remaining_samples,
            # TODO: Check where do we get this value from
            format = format,
            pts = None,
            time_base = None
        )
        
        frames.append(
            AudioFrameWrapped(
                frame = silent_frame,
                is_from_empty_part = True
            )
        )

    return frames

def get_videoframe_copy(
    video_frame: VideoFrame
) -> 'VideoFrame':
    """
    Create a copy of the provided VideoFrame
    instance. This is similar to .copy().
    """
    return VideoFrame.from_ndarray(
        array = video_frame.to_ndarray(format = video_frame.format.name),
        format = video_frame.format.name
    )

# TODO: I think we have this method in one
# of our numpy libraries
def remove_alpha_channel(
    array: 'np.ndarray'
) -> 'np.ndarray':
    """
    Remove the alpha channel from the given
    'array', if existing.
    """
    return (
        array[:, :, :3]
        if array.shape[-1] == 4 else
        array
    )


from yta_editor.media.abstract import _Media
from yta_editor.sources.abstract import _VideoSource
from yta_editor.sources.video import VideoFileSource, VideoImageSource, VideoColorSource, VideoNumpySource
from yta_editor.decorators import with_t_adjusted_to_media
from yta_editor.utils import apply_audio_effects_to_frame_at_t, apply_video_effects_to_frame_at_t
from yta_video_pyav.reader.filter.dataclass import GraphFilter
from yta_video_opengl.nodes import TimedNode
from av.video.frame import VideoFrame
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from typing import Union

    
class _VideoMedia(_Media):

    @property
    def size(
        self
    ) -> tuple[int, int]:
        """
        The size of the video frame, expressed
        as (width, height).
        """
        return self.source.size

    def __init__(
        self,
        source: _VideoSource,
        start: Union[int, float, Fraction] = 0.0,
        end: Union[int, float, Fraction, None] = None,
    ):
        super().__init__(
            source = source,
            start = start,
            end = end
        )

    def add_effect(
        self,
        effect: TimedNode
    ) -> 'Video':
        """
        Add the provided 'effect' to the video.
        """
        ParameterValidator.validate_mandatory_instance_of('effect', effect, 'TimedNode')

        self._effects.add_effect(effect)
        
        return self
    
    @with_t_adjusted_to_media
    def get_video_frame_at_t(
        self,
        t: Union[int, float, Fraction]
    ) -> Union[VideoFrame, None]:
        """
        Get the video frame with the given 't' time
        moment, using the video cache system.
        """
        print(f'Getting frame from {str(float(t + self.start))} that is actually {str(float(t))}')
        
        frame = self.source.get_video_frame_at_t(t)

        return (
            apply_video_effects_to_frame_at_t(
                effects_stack = self._effects,
                frame = frame,
                t = t
            )
            if frame is not None else
            None
        )
    
    @with_t_adjusted_to_media
    def get_audio_frames_at_t(
        self,
        t: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction, None] = None
    ):
        """
        Get the sequence of audio frames for the 
        given video 't' time moment, using the
        audio cache system.

        This is useful when we want to write a
        video frame with its audio, so we obtain
        all the audio frames associated to it
        (remember that a video frame is associated
        with more than 1 audio frame).
        """
        video_fps = (
            self.fps
            if video_fps is None else
            video_fps
        )

        print(f'Getting audio frames from {str(float(t + self.start))} that is actually {str(float(t))}')
        for frame in self.source.get_audio_frames_at_t(
            t = t,
            video_fps = video_fps
        ):
            yield apply_audio_effects_to_frame_at_t(
                effects_stack = self._effects,
                frame = frame,
                t = t
            )
    
class VideoFileMedia(_VideoMedia):
    """
    A video media that is read from a video
    file and can be subclipped to a specific
    time range.
    """

    @property
    def copy(
        self
    ) -> 'VideoFileMedia':
        """
        Get a copy of this instance with the same
        source, time range and effects.
        """
        copy = VideoFileMedia._init_with_source(
            source = self.source,
            start = self.start,
            end = self.end
        )

        copy._effects = self._effects.copy

        return copy

    @property
    def ticks_per_frame(
        self
    ) -> int:
        """
        The number of ticks per video frame. A
        tick is the minimum amount of time and
        is the way 'pts' is measured, in ticks.

        This means that the 'pts' value will
        be increased this amount from one video
        frame to the next one.

        How we obtain it:
        - `(1 / fps) / time_base`
        """
        return self.source.ticks_per_frame
    
    @property
    def duration(
        self
    ) -> Fraction:
        """
        The duration of the video, that can be
        shorter than the video file source
        duration if the user requested for it.

        The formula:
        - `self.end - self.start`
        """
        return self.end - self.start
    
    @property
    def number_of_frames(
        self
    ) -> Union[int, None]:
        """
        The number of frames of the video.
        """
        return self.source.number_of_frames
    
    @property
    def fps(
        self
    ) -> Union[Fraction, None]:
        """
        The frames per second of the video.
        """
        return self.source.fps
    
    @property
    def audio_fps(
        self
    ) -> Union[int, None]:
        """
        The frames per second of the audio.
        """
        return self.source.audio_fps
    
    @property
    def codec_name(
        self
    ) -> Union[str, None]:
        """
        The name of the codec.
        """
        return self.source.codec_name
    
    @property
    def pixel_format(
        self
    ) -> Union[str, None]:
        """
        The pixel format.
        """
        return self.source.pixel_format

    @property
    def size(
        self
    ) -> tuple[int, int]:
        """
        The size of the video frames expressed 
        like (width, height).
        """
        return self.source.size
    
    @property
    def width(
        self
    ) -> int:
        """
        The width of the video frames in pixels.
        """
        return self.size[0]
    
    @property
    def height(
        self
    ) -> int:
        """
        The height of the video frames in pixels.
        """
        return self.size[1]
    
    @property
    def time_base(
        self
    ) -> Union[Fraction, None]:
        """
        The time base of the video.
        """
        return self.source.time_base
    
    @property
    def audio_time_base(
        self
    ) -> Union[Fraction, None]:
        """
        The time base of the audio.
        """
        return self.source.audio_time_base

    def __init__(
        self,
        filename: str,
        start: Union[int, float, Fraction] = 0.0,
        end: Union[int, float, Fraction, None] = None,
        # These are ffmpeg filters
        video_filters: list[GraphFilter] = [],
        audio_filters: list[GraphFilter] = []
    ):
        super().__init__(
            source = VideoFileSource(
                filename = filename,
                video_filters = video_filters,
                audio_filters = audio_filters
            ),
            start = start,
            end = end
        )

    def add_video_filter(
        self,
        filter: GraphFilter
    ) -> 'VideoFileMedia':
        """
        Add a video filter to the list of filters
        to apply.
        """
        self.source.add_video_filter(filter)

        return self
    
    def add_audio_filter(
        self,
        filter: GraphFilter
    ) -> 'VideoFileMedia':
        """
        Add an audio filter to the list of filters
        to apply.
        """
        self.source.add_audio_filter(filter)

        return self

class VideoImageMedia(_VideoMedia):
    """
    A video media that is made by an static
    image file.
    """

    @property
    def copy(
        self
    ) -> 'VideoImageMedia':
        """
        Get a copy of this instance with the same
        source, time range and effects.
        """
        copy = VideoImageMedia._init_with_source(
            source = self.source,
            start = self.start,
            end = self.end
        )

        copy._effects = self._effects.copy

        return copy

    @property
    def filename(
        self
    ) -> str:
        """
        The filename of the original image.
        """
        return self.source.filename

    # TODO: Maybe rename this property (?)
    @property
    def do_include_alpha(
        self
    ) -> bool:
        """
        The internal flag to indicate if we
        want to consider the alpha channel or
        not.
        """
        return self.source._do_include_alpha

    def __init__(
        self,
        filename: str,
        duration: Union[int, float, Fraction],
        do_include_alpha: bool = True,
        size: Union[tuple[int, int], None] = None,
        # Need this to generate the silent audio frames
        audio_fps: int = 44_100,
        audio_samples_per_frame: int = 1024,
    ):
        """
        TODO: Maybe force size provided as parameter (?)
        """
        # We need to dynamically set the frame format
        # to accept (or not) the alpha channel
        frame_format = (
            'rgba'
            if do_include_alpha else
            'rgb24'
        )

        super().__init__(
            source = VideoImageSource(
                filename = filename,
                do_include_alpha = do_include_alpha,
                frame_format = frame_format,
                size = size,
                audio_fps = audio_fps,
                audio_samples_per_frame = audio_samples_per_frame
            ),
            start = 0,
            end = duration
        )

class VideoColorMedia(_VideoMedia):
    """
    A video media that is made with a static
    uniform color frame.
    """

    @property
    def copy(
        self
    ) -> 'VideoColorMedia':
        """
        Get a copy of this instance with the same
        source, time range and effects.
        """
        copy = VideoColorMedia._init_with_source(
            source = self.source,
            start = self.start,
            end = self.end
        )

        copy._effects = self._effects.copy

        return copy
    
    @property
    def color(
        self
        # TODO: Apply format
    ) -> any:
        """
        The color that will be used to make the
        frame that will be played its whole
        duration.
        """
        return self.source._color
    
    @property
    def size(
        self
    ) -> tuple[int, int]:
        """
        The size of the media frame.
        """
        return self.source.size

    def __init__(
        self,
        # TODO: Apply format
        color: any,
        duration: Union[int, float, Fraction],
        size: tuple[int, int] = (1920, 1080),
        transparency: Union[float, None] = 0.0,
        # TODO: Should we accept 'frame_format' (?)
        #frame_format: str = 'rgba',
        # Need this to generate the silent audio frames
        audio_fps: int = 44_100,
        audio_samples_per_frame: int = 1024,
    ):
        """
        TODO: The 'color' must be handled, by now
        we don't care about your argument x).
        """
        frame_format = (
            'rgba'
            if transparency is not None else
            'rgb24'
        )

        super().__init__(
            source = VideoColorSource(
                color = color,
                size = size,
                frame_format = frame_format,
                transparency = transparency,
                audio_fps = audio_fps,
                audio_samples_per_frame = audio_samples_per_frame
            ),
            start = 0,
            end = duration
        )

# TODO: Create 'VideoNumpyMedia'